package de.hft.stuttgart.citydoctor2.gui;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import de.hft.stuttgart.citydoctor2.datastructure.*;
import de.hft.stuttgart.citydoctor2.gui.filter.ViewFilter;
import de.hft.stuttgart.citydoctor2.math.Triangle3d;
import de.hft.stuttgart.citydoctor2.math.UnitVector3d;
import de.hft.stuttgart.citydoctor2.math.Vector3d;
import de.hft.stuttgart.citydoctor2.tesselation.TesselatedPolygon;
import javafx.scene.paint.Color;
import javafx.scene.paint.PhongMaterial;
import javafx.scene.shape.CullFace;
import javafx.scene.shape.DrawMode;
import javafx.scene.shape.MeshView;
import javafx.scene.shape.TriangleMesh;
import javafx.scene.shape.VertexFormat;

public class TriangulatedGeometry {

	private static final PhongMaterial GRID_MAT = new PhongMaterial(Color.BLACK);

	// random vector for calculating normal angles, for color determination
	private static final UnitVector3d AXIS = new Vector3d(19, 0.8, 1.5).normalize();

	private Vector3d movedBy;
	private List<MeshView> meshes;
	private List<PhongMaterial> materials;
	
	public static TriangulatedGeometry of(Geometry geom) {
		return of(geom.getPolygons());
	}
	
	public static TriangulatedGeometry of(Collection<? extends Polygon> polygons, Color basePolygonColor) {
		TriangulatedGeometry triGeom = new TriangulatedGeometry();
		triGeom.materials = new ArrayList<>();
		triGeom.meshes = new ArrayList<>();
		List<Vector3d> points = new ArrayList<>();
		for (Polygon p : polygons) {
            points.addAll(p.getExteriorRing().getVertices());
		}
		triGeom.movedBy = triGeom.findCenter(points);
		
		addPolygonDataToTriGeom(polygons, basePolygonColor, triGeom);
		return triGeom;
	}

	private static void addPolygonDataToTriGeom(Collection<? extends Polygon> polygons, Color basePolygonColor,
			TriangulatedGeometry triGeom) {
		for (Polygon p : polygons) {
			if (p instanceof CompositePolygon){
				continue;
			}
			TesselatedPolygon tp = p.tesselate();
			TriangleMesh triMesh = new TriangleMesh(VertexFormat.POINT_TEXCOORD);
			Map<Vector3d, Integer> indexMap = new HashMap<>();
			List<Vector3d> vertices = new ArrayList<>();
			int index = 0;
			for (Triangle3d t : tp.getTriangles()) {
				index = triGeom.filterDuplicates(triMesh, indexMap, vertices, index, t.getP1());
				index = triGeom.filterDuplicates(triMesh, indexMap, vertices, index, t.getP2());
				index = triGeom.filterDuplicates(triMesh, indexMap, vertices, index, t.getP3());
			}
			
			for (Vector3d point : vertices) {
				float x = (float) (point.getX() - triGeom.movedBy.getX());
				float y = (float) (point.getY() - triGeom.movedBy.getY());
				float z = (float) (point.getZ() - triGeom.movedBy.getZ());
				triMesh.getPoints().addAll(x, y, z);
			}
			triMesh.getTexCoords().addAll(0, 0);

			MeshView view = new MeshView(triMesh);
			view.setUserData(new PolygonClickDispatcher(p));

			PhongMaterial mat = triGeom.calculateMaterial(p, basePolygonColor);
			triGeom.materials.add(mat);
			triGeom.meshes.add(view);
		}
	}
	
	public static TriangulatedGeometry of(CityDoctorModel model, List<ViewFilter> filters) {
		List<Vector3d> points = new ArrayList<>();
		addPointsFromBuildings(model.getBuildings(), points);
		addPointsFromCityObject(model.getBridges(), points);
		addPointsFromCityObject(model.getLand(), points);
		addPointsFromCityObject(model.getTransportation(), points);
		addPointsFromCityObject(model.getVegetation(), points);
		addPointsFromCityObject(model.getWater(), points);
		
		TriangulatedGeometry triGeom = new TriangulatedGeometry();
		triGeom.materials = new ArrayList<>();
		triGeom.meshes = new ArrayList<>();
		triGeom.movedBy = triGeom.findCenter(points);
		
		addPolygonDataFromBuildings(model.getBuildings(), triGeom, filters);
		addPolygonDataFromCityObjects(model.getBridges(), triGeom, Color.CORAL, filters);
		addPolygonDataFromCityObjects(model.getLand(), triGeom, Color.BROWN, filters);
		addPolygonDataFromCityObjects(model.getTransportation(), triGeom, Color.YELLOW, filters);
		addPolygonDataFromCityObjects(model.getVegetation(), triGeom, Color.LIGHTGREEN, filters);
		addPolygonDataFromCityObjects(model.getWater(), triGeom, Color.LIGHTSKYBLUE, filters);
		addPolygonDataFromCityObjects(model.getCityFurniture(), triGeom, Color.BLUEVIOLET, filters);
		addPolygonDataFromCityObjects(model.getGenericCityObjects(), triGeom, Color.DARKGRAY, filters);
		
		return triGeom;
	}
	
	private static void addPolygonDataFromBuildings(List<Building> buildings, TriangulatedGeometry triGeom, List<ViewFilter> filters) {
		for (Building b : buildings) {
			addPolygonData(b, triGeom, Color.WHITE, filters);
			addPolygonDataFromBoundarySurfaces(b.getBoundarySurfaces(), triGeom, filters);
			addPolygonDataFromCityObjects(b.getBuildingInstallations(), triGeom, Color.WHITE, filters);
			for (Installation bi : b.getBuildingInstallations()) {
				addPolygonDataFromCityObjects(bi.getBoundarySurfaces(), triGeom, Color.WHITE, filters);
			}
			for (BuildingPart bp : b.getBuildingParts()) {
				addPolygonData(bp, triGeom, Color.WHITE, filters);
				addPolygonDataFromBoundarySurfaces(bp.getBoundarySurfaces(), triGeom, filters);
				addPolygonDataFromCityObjects(bp.getBuildingInstallations(), triGeom, Color.WHITE, filters);
			}
		}
	}

	private static void addPolygonDataFromBoundarySurfaces(List<BoundarySurface> boundarySurfaces,
			TriangulatedGeometry triGeom, List<ViewFilter> filters) {
		for (BoundarySurface bs : boundarySurfaces) {
			addPolygonData(bs, triGeom, Color.WHITE, filters);
			for (Opening o : bs.getOpenings()) {
				addPolygonData(o, triGeom, Color.WHITE, filters);
			}
		}
	}

	private static void addPolygonDataFromCityObjects(List<? extends CityObject> cos,
			TriangulatedGeometry triGeom, Color color, List<ViewFilter> filters) {
		for (CityObject co : cos) {
			addPolygonData(co, triGeom, color, filters);
		}
	}

	private static void addPolygonData(CityObject co, TriangulatedGeometry triGeom, Color color, List<ViewFilter> filters) {
		for (Geometry geom : co.getGeometries()) {
			if (isGeometryFiltered(co, geom, filters)) {
				continue;
			}
			List<Polygon> polygons = new ArrayList<>();
			for (Polygon p : geom.getPolygons()) {
				if (p.isLink()) {
					continue;
				}
				polygons.add(p);
			}
			addPolygonDataToTriGeom(polygons, color, triGeom);
		}
	}
	
	private static boolean isGeometryFiltered(CityObject co, Geometry geom, List<ViewFilter> filters) {
		for (ViewFilter filter : filters) {
			if (filter.allowedToUse(co, geom)) {
				return false;
			}
		}
		return true;
	}

	private static void addPointsFromBuildings(List<Building> buildings, List<Vector3d> points) {
		for (Building b : buildings) {
			addPoints(b, points);
			for (BuildingPart bp : b.getBuildingParts()) {
				addPoints(bp, points);
				addPointsFromCityObject(bp.getBoundarySurfaces(), points);
				addPointsFromCityObject(bp.getBuildingInstallations(), points);
			}
			addPointsFromCityObject(b.getBoundarySurfaces(), points);
			addPointsFromCityObject(b.getBuildingInstallations(), points);
		}
	}

	private static void addPointsFromCityObject(List<? extends CityObject> cos, List<Vector3d> points) {
		for (CityObject co : cos) {
			addPoints(co, points);
		}
	}

	private static void addPoints(CityObject co, List<Vector3d> points) {
		for (Geometry geom : co.getGeometries()) {
			for (Polygon p : geom.getPolygons()) {
				if (p.isLink()) {
					continue;
				}
                points.addAll(p.getExteriorRing().getVertices());
			}
		}
	}

	public static TriangulatedGeometry of(Collection<? extends Polygon> polygons) {
		return of(polygons, Color.WHITE);
	}

	private int filterDuplicates(TriangleMesh triMesh, Map<Vector3d, Integer> indexMap, List<Vector3d> vertices,
			int index, Vector3d v) {
		Integer vertexIndex = indexMap.get(v);
		if (vertexIndex == null) {
			indexMap.put(v, index);
			vertices.add(v);
			vertexIndex = index;
			index++;
		}
		triMesh.getFaces().addAll(vertexIndex, 0);
		return index;
	}

	private PhongMaterial calculateMaterial(Polygon p, Color baseColor) {
		Vector3d normal = p.calculateNormalNormalized();

		BoundarySurface bs = p.getPartOfSurface();
		if (bs != null) {
			if (bs.getType() == BoundarySurfaceType.ROOF) {
				baseColor = Color.RED;
			} else if (bs.getType() == BoundarySurfaceType.GROUND) {
				baseColor = Color.KHAKI;
			}
		}
		
		baseColor = determineColorDependingOnParentType(p, baseColor);

		double cos = normal.dot(AXIS);
		double acos = Math.acos(cos);
		// normalize to range [0.3, 0.9]
		acos = acos / Math.PI;
		acos = acos * 0.6 + 0.3;

		Color derivedColor = baseColor.deriveColor(0, 1.0, acos, 1.0);
		return new PhongMaterial(derivedColor);
	}

	private Color determineColorDependingOnParentType(Polygon p, Color baseColor) {
		p = p.getOriginal();
		Polygon p1 = p.getLinkedFromPolygon();
		baseColor = changeBaseColorIfPolygonHasOpeningParent(p1, baseColor);
		baseColor = changeBaseColorIfPolygonHasOpeningParent(p, baseColor);
		return baseColor;
	}

	private Color changeBaseColorIfPolygonHasOpeningParent(Polygon p, Color baseColor) {
		if (p == null) {
			return baseColor;
		}
		CityObject parent = p.getParent().getParent();
		if (parent instanceof Opening op) {
			if (op.getType() == OpeningType.DOOR) {
				baseColor = Color.ORANGE;
			} else {
				baseColor = Color.TEAL;
			}
		}
		return baseColor;
	}

	private Vector3d findCenter(List<Vector3d> points) {
		double xMin = Double.MAX_VALUE;
		double yMin = Double.MAX_VALUE;
		double zMin = Double.MAX_VALUE;

		double xMax = Double.NEGATIVE_INFINITY;
		double yMax = Double.NEGATIVE_INFINITY;
		double zMax = Double.NEGATIVE_INFINITY;

		for (Vector3d point : points) {
			if (point.getX() < xMin) {
				xMin = point.getX();
			}
			if (point.getX() > xMax) {
				xMax = point.getX();
			}
			if (point.getY() < yMin) {
				yMin = point.getY();
			}
			if (point.getY() > yMax) {
				yMax = point.getY();
			}
			if (point.getZ() < zMin) {
				zMin = point.getZ();
			}
			if (point.getZ() > zMax) {
				zMax = point.getZ();
			}
		}

		// center
		double x = (xMax - xMin) / 2 + xMin;
		double y = (yMax - yMin) / 2 + yMin;
		double z = (zMax - zMin) / 2 + zMin;
		return new Vector3d(x, y, z);
	}

	public Vector3d getMovedBy() {
		return movedBy;
	}

	public void setCullFace(CullFace currentCulling) {
		for (MeshView mesh : meshes) {
			mesh.setCullFace(currentCulling);
		}
	}

	public void setDrawMode(DrawMode currentDrawMode) {
		if (currentDrawMode == DrawMode.LINE) {
			for (MeshView mesh : meshes) {
				mesh.setDrawMode(currentDrawMode);
				mesh.setMaterial(GRID_MAT);
			}
		} else if (currentDrawMode == DrawMode.FILL) {
			for (int i = 0; i < meshes.size(); i++) {
				MeshView mesh = meshes.get(i);
				mesh.setDrawMode(currentDrawMode);
				mesh.setMaterial(materials.get(i));
			}
		}
	}

	public List<MeshView> getMeshes() {
		return meshes;
	}
}
